// ****************************************************************************
//
// bt-gasel.mps        gasel for Group or Area SELector
//
// 2017.01.11   v1.0   First release
// 2017.01.12   v1.0a  Converted all filenames to lowercase
//                     // thx xqtr for suggesting this as target for multi-system
//                     // compatibility
// 2017.01.24   v1.0b  Files structure externalized in reusable autonomous file
// 2017.01.26   v1.1a  Groups/areas count limit raised at 30,000, using dynamic buffer
//                     Second big code rewrite and reorganization, less memory used,
//                     less arrays manipulations, search rewrite (1st stage)
// 2017.02.02   v1.1b  sys_functions.inc was missing from zip package
// 2017.02.03   v1.1c  Bugfix on PageUp (1st time moves to top of page, then to 1st
// 2018.04.23   v1.1d  Bugfix for MysticBBS v1.12 A39 (data structures changed)
//                     sys_recs.inc updated | special thanks to Necromaster@Necronomicon BBS
//                     for providing missed specs
//
// Dynamic Lightbar menu for Groups/Areas selection, both Messages and Files
// customizable with optional config file that can be passed in the menu config
//
// Based on XQ-FMAG.ZIP by xqtr from Another Droid BBS :
//
//    - complete rewrite
//    - code optimization and clarification
//    - customization through external config file
//    - one single MPS for all 4 functions (Files/Messages Groups/Areas)
//    - default values if no custom config file provided
//    - display file for each function customizable through config file
//    - single procedure Draw_Bar instead of DrawON et DrawOFF
//    - uses of Case tests instead of nested/unnested if then else
//    - added support for Right/Left arrows (equiv of PgUp/PgDn)
//    - global code comment and clean up
//    - bugfix of Base/Group sorted/delete/added and displayed list order don't match
//      internal Unique IDs
//
// This Group/Area SELector with lightbar supports up to 30,000 areas per groups or
// up to 30,000 groups and then up to 30,000 for each group. That should fit most need.
//
// To achieve this, I'm using a array of all Groups or Areas (globally called "zones")
// which is arRecNos, an array of words that stores the sequential number of all valid
// zones from the given .dat file (filebase, message base, or groups). This is limited
// to 30,000 items due to MPL memory implementation (64k max per array).
//
// Then, I load a buffer which is a subset of this full array of sequentials IDs limited
// to csiBuffer_Size elements with zone name, unique ID and sequential number. You may
// set to csiBuffer_Size to whatever you like below 1395 which is again the 64k limit of
// MPL.
//
// The visible list where you can move with arrows, home/end keys and do some search
// is maintained inside this buffer, and the buffer is reloaded when needed (moving
// past what is already loaded, or before what is already loaded).
//
// On small or slow system, keeping this csiBuffer_Size constant pretty low should give
// better results since less data is read each time (but they're read more often).
//
// I keep it in the 200-400 area and process time is invisible (MacOS, Core i5).
//
//  Coming development in future releases :
//
//    - better search (more like a filter than a start listing at position x)
//    - multi-column through config file also
//    - any ideas ? send them please :)
//
// Redesigned & Rewritten for BackToTheRoots BBS by Fabian
//            telnet : BackToTheRoots.it-awareness.ch:2332
//                                email : fabian@lucchi.me
// ****************************************************************************

Uses CFG
Uses MGroup
Uses FGroup
Uses FBase
Uses MBase
Uses User

include sys_recs.inc

Type
  TInternalzone   = Record
                      sName       : String[40];               // Keep the biggest (base=40, group=30)
                      liUniqueID  : LongInt;                  // Internal Mystic unique ID that doesn't change, ever
                      wSeq        : Word;                     // Array position in arRecNO
                    End;

Type
  TBufferState    = Record
                      liFirst     : LongInt;                  // First zone loaded, sequential # of valid zones
                      liInFile    : LongInt;                  // Number of zones available in .dat file
                      liCount     : Word;                  // Number of zones loaded
                    End;

Const
  cssMPL_Name                       = 'bt-gasel';
  cssMPL_Version                    = '1.1d';

Const
  csiDefZone_Type                   =  1;                     // Default to File Groups
  csiDefPos_StartX                  = 21;                     // Left column of usable space to display list
  csiDefPos_StartY                  =  9;                     // Top row of usable space to display list
  csiDefPos_EndX                    = 39;                     // Right column of usble space to diplay list
  csiDefPos_EndY                    = 20;                     // Botton row of usable space to display list
  csiDefPos_SearchX                 = 20;                     // Column to display search string
  csiDefPos_SearchY                 = 24;                     // Row to display search string

  cssDefColors_Search_Prefix        = '|15';                  // MCI codes to send before displaying the search string
  cssDefColors_Bar_On_Prefix        = '|15|17';               // MCI codes to send before drawing an highlighted bar
  cssDefColors_Bar_On_Suffix        = '|16';                  // MCI codes to send after drawing an highlighted bar
  cssDefColors_Bar_Off_Prefix       = '|07';                  // MCI codes to send before drawing a dimmed bar

  cssDefDisplay_File                = 'gasel-fg';             // Default file to display as background
  cssDefHelp_File                   = 'gaselhlp';             // Default help file to display

  csiBuffer_Size                    = 300;                    // Max number of zones to load in memory at once
  csiIndexTable_Size                = 30000;                  // Maximum number of zones/areas in one particular group

  cssParameters_ZoneType            = 'Zone_Type';            // Zone type (1=FG, 2=FA, 3=MG, 4=MA)
  cssParameters_TopLeftX            = 'Top_Left_X';           // Param name for left column of list
  cssParameters_TopLeftY            = 'Top_Left_Y';           // Param name for top row of list
  cssParameters_BottomRightX        = 'Bottom_Right_X';       // Param name for right colum of list
  cssParameters_BottomRightY        = 'Bottom_Right_Y';       // Param name for botton row of list
  cssParameters_DisplayFile         = 'Display_File';         // Param name for display file to use
  cssParameters_HelpFile            = 'Help_File';            // Param name for help file to be displayed
  cssParameters_SearchX             = 'Pos_Search_X';         // Param name for column of search string
  cssParameters_SearchY             = 'Pos_Search_Y';         // Param name for row of search string
  cssParameters_Search_Prefix       = 'Search_Color_Prefix';  // Param name for MCI before search string
  cssParameters_Bar_On_Prefix       = 'Bar_HighLight_Prefix'; // Param name for MCI before highlighted bar
  cssParameters_Bar_On_Suffix       = 'Bar_HighLight_Suffix'; // Param name for MCI after highlighted bar
  cssParameters_Bar_Off_Prefix      = 'Bar_Dimmed_Prefix';    // Param name for MCI before dimmed bar

  csiZone_Groups_File               = 1;
  csiZone_Bases_File                = 2;
  csiZone_Groups_Message            = 3;
  csiZone_Bases_Message             = 4;

  cssMenu_FG                        = 'FG';
  cssMenu_FA                        = 'FA';
  cssMenu_MG                        = 'MG';
  cssMenu_MA                        = 'MA';

Var
  byZone_Type               : Byte = csiDefZone_Type;
  byPos_StartX              : Byte = csiDefPos_StartX;
  byPos_StartY              : Byte = csiDefPos_StartY;
  byPos_EndX                : Byte = csiDefPos_EndX;
  byPos_EndY                : Byte = csiDefPos_EndY;
  byTableWidth              : Byte;
  byPos_SearchX             : Byte = csiDefPos_SearchX;
  byPos_SearchY             : Byte = csiDefPos_SearchY;
  sColor_Search_Prefix      : String  = cssDefColors_Search_Prefix;
  sColor_Bar_On_Prefix      : String  = cssDefColors_Bar_On_Prefix;
  sColor_Bar_On_Suffix      : String  = cssDefColors_Bar_On_Suffix;
  sColor_Bar_Off_Prefix     : String  = cssDefColors_Bar_Off_Prefix;
  sDisplayFile              : String  = cssDefDisplay_File;
  sHelpFile                 : String  = cssDefHelp_File;

  arRecNos                  : Array[1..csiIndexTable_Size] Of Word;
  wRecNo_Count              : Word;
  arZones                   : Array[1..csiBuffer_Size] Of TInternalZone;
  recGroup                  : TGroupBase;
  recFile                   : TFileBase;
  recMessage                : TMessageBase;
  wRecSize                  : Word;                              // Record size, depending on selected Zone_Type
  sFilename                 : String;                            // Filename to read data from, depends on selected Zone_Type
  liSeq                     : LongInt;                           // Count sequential numbers while loading zones
  liRec                     : LongInt;                           // Trace for current record #

  liNextSeqToLoad           : LongInt;                           // LongInt required to prevent overflow during calc of next/previous zone
  wPreviousSeq              : Word;

  byTemp                    : Byte;

  liZoneNumberOf1stRow      : LongInt;                           // self explanatory, LongInt to prevent Word overflow (0 - 1 = 65535)
  liHighlightedZone         : LongInt;                           // Zone highlighted (array# in current buffer)
  bDone                     : Boolean;                           // Is a select or esc issued ?
  cChar                     : Char;
  cChar2                    : Char;
  sSearch                   : String;                            // Search string
  wSearch                   : Word;                              // Position of corresponding zone in array
  wCount                    : Word;
  byTableLength             : Byte;                              // Height of list (bottom row - top row)
  sDefaultMenu              : String;                            // Default Mystic menu to call if any problem

  bsBuffer                  : TBufferState;                      // For dynamic buffer, actual slice of data of cssBuffer_Size lines
  bFirstRun                 : Boolean = True;                    // If true, read the data file until to end to count valid zones
                                                                 // and store it on bsBuffer.liInFile / this way we do not parse file
                                                                 // each time for nothing and no useless read either

// Default/Generic/multi purpose functions
include sys_functions.inc

//
// ***************************************************************************
// Clear search string & display
// ***************************************************************************
//

Procedure ClearSearch;
Begin
  GotoXY(byPos_SearchX, byPos_SearchY);
  Write (StrRep(' ', Length(sSearch)));
  sSearch := '';
End; // ClearSearch;

//
// ***************************************************************************
// Generic Draw Bar procedure
// >iPos = row of bar to draw
// >bOn  = is bar hightlighted
// ***************************************************************************
//

Procedure Draw_Bar(wPos : Integer; bOn : Boolean);
Begin
  GotoXY (byPos_StartX, byPos_StartY + wPos - liZoneNumberOf1stRow);
  Case bOn Of
    True  : Begin
              Write(sColor_Bar_On_Prefix+' ' + PadRT(StripMCI(arZones[wPos].sName), byTableWidth, ' ') + sColor_Bar_On_Suffix);
            End;
    False : Begin
              Write(sColor_Bar_Off_Prefix+' ' + PadRT(arZones[wPos].sName, byTableWidth, ' '));
            End;
  End; // Case bOn ....
End;

//
// ***************************************************************************
// Redraw full page
// ***************************************************************************
//

Procedure DrawFullPage;
Var
  wLocalPos : Word;
Begin
  GotoXY(20,1);
  wLocalPos := liHighlightedZone;

  For byTemp := 0 to byTableLength Do Begin
    liHighlightedZone := liZoneNumberOf1stRow + byTemp;
    Draw_Bar(liHighlightedZone, False);
  End; // For byTemp := 0 To ...
  liHighlightedZone   := wLocalPos;
  Draw_Bar(liHighlightedZone, True);
End; // Procedure DrawFullPage

//
// ***************************************************************************
// Find Seq # in array
// ***************************************************************************
//
Function GetSeqPosInBuffer(wSequential : Word) : Word;
Var
  wLocal : Word;
Begin
  GetSeqPosInBuffer := 0;

  For wLocal := 1 To csiBuffer_Size Do Begin
    If arZones[wLocal].wSeq = wSequential Then Begin
      GetSeqPosInBuffer := wLocal;
      Break;
    End; // If arZones[wLocal].wSeq = wSequential
  End; // For wLocal
End; // Function GetSeqPosInBuffer

//
// ***************************************************************************
// Find liUniqueID # in array
// ***************************************************************************
//
Function GetUniqueIDPosInBuffer(liUnique : LongInt) : Word;
Var
  wLocal : Word;
Begin
  GetUniqueIDPosInBuffer := 0;

  For wLocal := 1 To csiBuffer_Size Do Begin
    If arZones[wLocal].liUniqueID = liUnique Then Begin
      GetUniqueIDPosInBuffer := wLocal;
      Break;
    End; // If arZones[wLocal].liUniqueID = liUnique
  End; // For wLocal
End; // Function GetUniqueIDPosInBuffer

//
// ***************************************************************************
// Load RecNos of valid zones in permanent array, once
// ***************************************************************************
//

Procedure LoadValidZones;
Var
  fZones     : File;
  wLocal     : Word;
Begin
  If Not(FileExist(CfgDataPath + sFileName)) Then Begin
    StopWithMessage('No system file ('+CfgDataPath+sFilename+')')
  End; // If Not(FileExist...

  fAssign(fZones, CfgDataPath + sFilename, 66);
  fReset(fZones);

  liSeq        := 0;
  liRec        := 0;
  wRecNo_Count := 0;

  // Read all zones
  Repeat
    liRec := liRec + 1;
    Case byZone_Type Of
      csiZone_Groups_File,
      csiZone_Groups_Message  : Begin
                                  fReadRec(fZones, recGroup);
                                  If ACS(recGroup.sACS) Then Begin
                                    liSeq                                := liSeq + 1;
                                    arRecNos[liSeq]                      := liRec;
                                  End; // If ACS()
                                End;
      csiZone_Bases_File      : Begin
                                  fReadRec(fZones, recFile);
                                  If ACS(recFile.sListACS) Then Begin
                                    liSeq                                := liSeq + 1;
                                    ArRecNos[liSeq]                      := liRec;
                                  End; // If ACS
                                End;
      csiZone_Bases_Message   : Begin
                                  fReadRec(fZones, recMessage);
                                  If ACS(recMessage.sListACS) Then Begin
                                    liSeq                                := liSeq + 1;
                                    arRecNos[liSeq]                      := liRec;
                                  End; // If ACS
                                End;
      Else StopWithMessage('No Zone Type selected !!! check .cfg file');
    End; // case iZone_Type ...
  Until fEOF(fZones) Or (liSeq >= csiIndexTable_Size-1); // Limit data read to buffer size
  wRecNo_Count := liSeq;

  // Cleaning end of array
  For wLocal := liSeq+1 To csiIndexTable_Size Do Begin
    arRecNos[wLocal] := 0;
  End; // For wLocal

End; // Procedure LoadValidZones;

//
// ***************************************************************************
// Generic Load Zones
// ***************************************************************************
//

Procedure LoadBuffer(wStartingFrom : LongInt);
Var
  fZones        : File;
  wRecNoIndex   : Word;
  wBufferIndex  : Word;
Begin
  fAssign(fZones, CfgDataPath + sFilename, 66);
  fReset(fZones);

  wRecNoIndex  := wStartingFrom;
  wBufferIndex :=1;

  Repeat
    // Move to required RecNo
    fSeek(fZones, (arRecNos[wRecNoIndex] - 1) * wrecSize);

    Case byZone_Type Of
      csiZone_Groups_File,
      csiZone_Groups_Message  : Begin
                                  fReadRec(fZones, recGroup);

                                  arZones[wBufferIndex].sName      := recGroup.sName;
                                  arZones[wBufferIndex].liUniqueID := recGroup.iUniqueID;
                                  arZones[wBufferIndex].wSeq      := wRecNoIndex;
                                End;
      csiZone_Bases_File      : Begin
                                  fReadRec(fZones, recFile);
                                  arZones[wBufferIndex].sName      := recFile.sName;
                                  arZones[wBufferIndex].liUniqueID := recFile.wIndex;
                                  arZones[wBufferIndex].wSeq      := wRecNoIndex;
                                End;
      csiZone_Bases_Message   : Begin
                                  fReadRec(fZones, recMessage);
                                  arZones[wBufferIndex].sName      := recMessage.sName;
                                  arZones[wBufferIndex].liUniqueID := recMessage.wIndex;
                                  arZones[wBufferIndex].wSeq      := wRecNoIndex;
                                End;
      Else StopWithMessage('No Zone Type selected !!! check .cfg file');
    End; // case byZone_Type ...

    wRecNoIndex  := wRecNoIndex + 1;
    wBufferIndex := wBufferIndex + 1;
  Until fEOF(fZones) Or (wBufferIndex >= csiBuffer_Size) Or (wRecNoIndex > wRecNo_Count); // Limit data read to buffer size

  fClose(fZones);

  bsBuffer.liCount := wBufferIndex-1;
End; // Procedure LoadBuffer

//
// ***************************************************************************
// Load buffer but the starting Seq# is the first string match
// when found, LoadBuffer with normal procedure and position display accordingly
// sSearchString <-  string to search
// integer       ->  liUniqueID of first match or 0 if not found
// ***************************************************************************
//

Procedure SearchWithString(sSearchString : String);
Var
  fZones        : File;
  wRecNoIndex   : Word;
  wLen          : Word;
  bFound        : Boolean = False;
  liIDToSearch  : LongInt;
  liNextSeq     : LongInt;
Begin
  fAssign(fZones, CfgDataPath + sFilename, 66);
  fReset(fZones);

  wRecNoIndex  := 0;
  wLen         := Length(sSearchString);

  // We read every zone until a string match on the zone name
  Repeat
    wRecNoIndex := wRecNoindex + 1;

    // Move to required RecNo
    fSeek(fZones, (arRecNos[wRecNoIndex] - 1) * wrecSize);

    Case byZone_Type Of
      csiZone_Groups_File,
      csiZone_Groups_Message  : Begin
                                  fReadRec(fZones, recGroup);

                                  If Upper(Copy(recGroup.sName, 1, wLen)) = sSearchString Then Begin
                                    // Found, we load from here
                                    liIDToSearch := recGroup.iUniqueID;
                                    bFound := True;
                                  End; // If Copy(recGroup.sName, 1, wLen)...
                                End;
      csiZone_Bases_File      : Begin
                                  fReadRec(fZones, recFile);

                                  If Upper(Copy(recFile.sName, 1, wLen)) = sSearchString Then Begin
                                    liIDToSearch := recFile.wIndex;
                                    bFound := True;
                                  End
                                  Else Begin
                                  End; // If Copy(recFile.sName, 1, wLen)...
                                End;
      csiZone_Bases_Message   : Begin
                                  fReadRec(fZones, recMessage);

                                  If Upper(Copy(recMessage.sName, 1, wLen)) = sSearchString Then Begin
                                    liIDToSearch := recMessage.wIndex;
                                    bFound := True;
                                  End; // If Copy(recFile.sName, 1, wLen)...
                                End;
      Else StopWithMessage('No Zone Type selected !!! check .cfg file');
    End; // case byZone_Type ...
  Until bFound Or fEOF(fZones) Or (wRecNoIndex > wRecNo_Count); // Limit data read to buffer size
  fClose(fZones);

  If bFound Then Begin
    // We found data, load buffer and return control for display
    // If actual position is farther that buffer capacity, load a full buffer
    If wRecNo_Count - wRecNoIndex < csiBuffer_Size - 1 Then Begin
      liNextSeq := wRecNo_Count - csiBuffer_Size + 2;
      If liNextSeq < 1 Then liNextSeq := 1;
    End
    Else Begin
      liNextSeq := wRecNoIndex;
    End;

    LoadBuffer(liNextSeq);

    wRecNoIndex := GetUniqueIDPosInBuffer(liIDToSearch);

    If bsBuffer.liCount - wRecNoIndex < byTableLength Then Begin
      // item is on the last page, move there and position bar accordingly
      liZoneNumberOf1stRow := bsBuffer.liCount - byTableLength;
      If liZoneNumberOf1stRow < 1 Then liZoneNumberOf1stRow := 1;
      liHighlightedZone    := wRecNoIndex;
    End
    Else Begin
      // Item is in the middle of the buffer, display on first line
      liZoneNumberOf1stRow := wRecNoIndex;
      liHighlightedZone    := liZoneNumberOf1stRow;
    End; // If bsBuffer.liCount
    DrawFullPage;
  End
  Else Begin
    // Nothing found, signal to calling proc
  End; // If bFound
End; // Procedure SearchWithString

//
// ***************************************************************************
// We load any custom parameters if required
// ***************************************************************************
//

Procedure Load_Parameters;
Var
  tfInput  : File;
  sLine    : String;
  sParam   : String;

  //
  // We extract the value (string) for a given parameter name (as read from file)
  //
  Function ExtractParamValue(sString, sName : String) : String;
  Begin
    ExtractParamValue := StripB(Copy(sString, Pos('=', sString)+1, 255), ' ');
  End;

Begin
  //
  // If Parameters File Doesn't Exists, We Use Default Values Through
  // Default Init At Variables Declaration
  //
  If FileExist(CfgMpePath + ParamStr(1)) Then Begin
    // Open Parameters File For Customized Settings only if file does exist
    fAssign(tfInput, CfgMpePath + ParamStr(1), 66);
    fReset(tfInput);

    Repeat
      fReadLn(tfInput, sLine);

      // Do not process comments in that file
      If Copy(sLine, 1, 2)<>'//' Then Begin
        If Pos(cssParameters_ZoneType, sLine)>0 Then Begin
          sParam := ExtractParamValue(sLine, cssParameters_ZoneType);
          byZone_Type := Str2Int(sParam);
        End
        Else If Pos(cssParameters_TopLeftX, sLine)>0 Then Begin
          sParam := ExtractParamValue(sLine, cssParameters_TopLeftX);
          byPos_StartX := Str2Int(sParam);
        End
        Else If Pos(cssParameters_TopLeftY, sLine)>0 Then Begin
          sParam := ExtractParamValue(sLine, cssParameters_TopLeftY);
          byPos_StartY := Str2Int(sParam);
        End
        Else If Pos(cssParameters_BottomRightX, sLine)>0 Then Begin
          sParam := ExtractParamValue(sLine, cssParameters_BottomRightX);
          byPos_EndX := Str2Int(sParam);
        End
        Else If Pos(cssParameters_BottomRightY, sLine)>0 Then Begin
          sParam := ExtractParamValue(sLine, cssParameters_BottomRightY);
          byPos_EndY := Str2Int(sParam);
        End
        Else If Pos(cssParameters_DisplayFile, sLine)>0 Then Begin
          sDisplayFile := ExtractParamValue(sLine, cssParameters_DisplayFile);
        End
        Else If Pos(cssParameters_SearchX, sLine)>0 Then Begin
          sParam := ExtractParamValue(sLine, cssParameters_SearchX);
          byPos_SearchX := Str2Int(sParam);
        End
        Else If Pos(cssParameters_SearchY, sLine)>0 Then Begin
          sParam := ExtractParamValue(sLine, cssParameters_SearchY);
          byPos_SearchY := Str2Int(sParam);
        End
        Else If Pos(cssParameters_Search_Prefix, sLine)>0 Then Begin
          sColor_Search_Prefix := ExtractParamValue(sLine, cssParameters_Search_Prefix);
        End
        Else If Pos(cssParameters_Bar_On_Prefix, sLine)>0 Then Begin
          sColor_Bar_On_Prefix := ExtractParamValue(sLine, cssParameters_Bar_On_Prefix);
        End
        Else If Pos(cssParameters_Bar_On_Suffix, sLine)>0 Then Begin
          sColor_Bar_On_Suffix := ExtractParamValue(sLine, cssParameters_Bar_On_Suffix);
        End
        Else If Pos(cssParameters_Bar_Off_Prefix, sLine)>0 Then Begin
          sColor_Bar_Off_Prefix := ExtractParamValue(sLine, cssParameters_Bar_Off_Prefix);
        End
        Else If Pos(cssParameters_HelpFile, sLine)>0 Then Begin
          sHelpFile := ExtractParamValue(sLine, cssParameters_HelpFile);
        End;
      End; // If Copy(sLine, 1, 2)....
    Until fEOF(tfInput);

    fClose(tfInput);

    byTableWidth := byPos_EndX - byPos_StartX;

    // Set default menu action if we need to fallback to Mystic (zone too big or no zone, or no ANSI)
    // Set data file to use depending on Zone-Type
    // Set record size depending on Zone-Type
    Case byZone_Type Of
      1 : Begin
            sDefaultMenu := cssMenu_FG;
            sFilename    := cssMystic_System_FG;
            wRecSize     := SizeOf(recGroup);
          End;
      2 : Begin
            sDefaultMenu := cssMenu_FA;
            sFilename    := cssMystic_System_FB;
            wRecSize     := SizeOf(recFile);
          End;
      3 : Begin
            sDefaultMenu := cssMenu_MG;
            sFileName    := cssMystic_System_MG;
            wRecSize     := SizeOf(recGroup);
          End;
      4 : Begin
            sDefaultMenu := cssMenu_MA;
            sFileName    := cssMystic_System_MB;
            wRecSize     := SizeOf(recMessage);
          End;
    End; // Case
  End; // If Not(FileExist)...
End; // Procedure Load_Parameters
//
// **************************************************************
// ***                Start of main program                   ***
// **************************************************************
//

Begin
  GetThisUser;

  // If no ANSI capabilities, use default Mystic function and exit immediately
  If Graphics = 0 Then Begin
    MenuCmd (sDefaultMenu, '');
    Halt;
  End;

  // Load Customized Parameters
  Load_Parameters;

  // Load Valid Zones in arRecNos
  wRecNo_Count      := 0;
  LoadValidZones;

  bsBuffer.liFirst  := 0;
  bsBuffer.liInFile := 0;
  bsBuffer.liCount  := 0;

  byTemp            := 1;

  sSearch           := '';
  wSearch           := 0;

  byTableLength := byPos_EndY-byPos_StartY;

  LoadBuffer(1);

  // If no zones defined, show default prompt and exit
  If bsBuffer.liCount = 0 Then begin
    WriteLn (GetPrompt(37));
    Halt;
  End; // If bsBuffer.liCount = 0

  // If zones count = max allowed zones, use appropriate builtin function and exit
  If bsBuffer.liCount = csiBuffer_Size Then Begin
    MenuCmd(sDefaultMenu, '');
    Halt;
  End; // If bsBuffer.liCount = ...

  // Show list template
  DispFile (sDisplayFile);

  liZoneNumberOf1stRow  := 1;
  liHighlightedZone     := 1;
  bDone                := False;

  // Initial First Page Draw
  DrawFullPage;

  Repeat
    cChar := ' ';

    If liZoneNumberOf1stRow+byTableLength < bsBuffer.liCount Then Begin
      cChar2 := Chr(245);
    End; // If (liZoneNumberOf1stRow + byTableLength)...

    cChar := ReadKey;

    If IsArrow Then Begin

      Case cChar Of
        Chr(71)     : Begin
                        // key HOME
                        ClearSearch;
                        liNextSeqToLoad      := 1;
                        liZoneNumberOf1stRow := liNextSeqToLoad;
                        liHighlightedZone    := liNextSeqToLoad;
                        LoadBuffer(liNextSeqToLoad);
                        DrawFullPage;
                      End; // Chr(71)

        Chr(79)     : Begin
                        // key END
                        ClearSearch;
                        
                        liNextSeqToLoad := wRecNo_Count - csiBuffer_Size;
                        If liNextSeqToLoad < 1 Then liNextSeqToLoad := 1; // less data than a single windows ?
                        LoadBuffer(liNextSeqToLoad);

                        If wRecNo_Count < byTableLength Then Begin
                          // Less data than a full page
                          liZoneNumberOf1stRow := 1;
                          liHighlightedZone    := bsBuffer.liCount;
                        End
                        Else Begin
                          // At least a full page, go to the end
                          liZoneNumberOf1stRow := bsBuffer.liCount - byTableLength;
                          liHighlightedZone    := bsBuffer.liCount;
                        End; // If wRecNo_Count
                        DrawFullPage;
                      End; // Chr(79)

        Chr(72)     : Begin
                        // key UP
                        ClearSearch;

                        If liHighlightedZone=1 Then Begin
                          // Top of page, check of more data is available
                          If arZones[1].wSeq = 1 Then Begin
                            // Top of list, do nothing, error message maybe ?
                            // TBD
                          End
                          Else Begin
                            liNextSeqToLoad          := arZones[1].wSeq + byTableLength - csiBuffer_Size + 1;
                            If liNextSeqToLoad < 1 Then Begin
                              liNextSeqToLoad        := 1;
                              wPreviousSeq           := arZones[1].wSeq;
                              LoadBuffer(liNextSeqToLoad);
                              liZoneNumberOf1stRow    := GetSeqPosInBuffer(wPreviousSeq - 1);
                              liHighlightedZone       := liZoneNumberOf1stRow;
                            End
                            Else Begin
                              liZoneNumberOf1stRow    := csiBuffer_Size - byTableLength - 1;
                              liHighlightedZone       := liZoneNumberOf1stRow;
                              LoadBuffer(liNextSeqToLoad);
                            End; // If liNextSeqToLoad
                            DrawFullPage;
                          End; // If liHighlightedZone = 1
                        End
                        Else If liHighlightedZone > liZoneNumberOf1stRow Then Begin
                          Draw_Bar(liHighlightedZone, False);
                          liHighlightedZone := liHighlightedZone - 1;
                          Draw_Bar(liHighlightedZone, True);
                        End
                        Else If liZoneNumberOf1stRow > 1 Then Begin
                          liZoneNumberOf1stRow  := liZoneNumberOf1stRow - 1;
                          liHighlightedZone     := liZoneNumberOf1stRow;
                          DrawFullPage;
                        End; // If liHighlightedZone > liZoneNumberOf1stRow...
                      End; // Chr(27)

        Chr(73),
        Chr(75)     : Begin
                        // key PageUp Or Left
                        ClearSearch;

                        // If not already on first line of screen, move there as first step
                        If liHighlightedZone <> liZoneNumberOf1stRow Then Begin
                          Draw_Bar(liHighlightedZone, False);
                          liHighlightedZone := liZoneNumberOf1stRow;
                          Draw_Bar(liHighlightedZone, True);
                        End
                        Else If liHighlightedZone - byTableLength < 1 Then Begin
                          // Buffer exhausted, we need to read more data from .dat files
                          wPreviousSeq    := arZones[liHighlightedZone].wSeq;
                          liNextSeqToLoad := wPreviousSeq - csiBuffer_Size;
                          If liNextSeqToLoad < 1 Then Begin
                            // End of data, position to load from top of file (first csiBuffer_Size records)
                            liNextSeqToLoad   := 1;
                            // Load only if new pos is <> previous pos (already at the end?)
                            If liNextSeqToLoad <> arZones[1].wSeq Then Begin
                              LoadBuffer(liNextSeqToLoad);
                            End
                            Else Begin
                            End; // If liNextSeqToLoad <> arZones[1].wSeq
                            liHighlightedzone := GetSeqPosInBuffer(wPreviousSeq) - byTableLength;
                            If liHighlightedZone < 1 Then Begin
                              liHighlightedZone := 1;
                            End;
                            liZoneNumberOf1stRow := liHighlightedZone;
                            DrawFullPage;
                          End
                          Else Begin
                            // Normal buffer read
                            LoadBuffer(liNextSeqToLoad + 1);
                            liZoneNumberOf1stRow := GetSeqPosInBuffer(wPreviousSeq - byTableLength) - 1;
                            liHighlightedZone    := liZoneNumberOf1stRow;
                            DrawFullPage;
                          End; // if liNextSeqToLoad > wRecNo_Count
                        End
                        // Buffer not exhausted, we move in it
                        Else Begin
                          // Normal move in buffer
                          liZoneNumberOf1stRow := liZoneNumberOf1stRow - byTableLength;
                          If liZoneNumberOf1stRow < 1 Then liZoneNumberOf1stRow := 1;
                          liHighlightedZone    := liZoneNumberOf1stRow;
                          DrawFullPage;
                        End; // If liZoneNumberOf1stRow+byTableLength...

                      End; // Chr(73)

        Chr(80)     : Begin
                        // key DOWN
                        ClearSearch;

                        If (liHighlightedZone = bsBuffer.liCount) Then Begin
                          // Is there more data to read ?
                          If arZones[bsBuffer.liCount].wSeq < wRecNo_Count Then Begin
                            // Yes, we can read more data
                            liNextSeqToLoad       := arZones[bsBuffer.liCount].wSeq - byTableLength +1;
                            liZoneNumberOf1stRow   := 1;
                            liHighlightedZone      := byTableLength + 1;

                            If liNextSeqToLoad + csiBuffer_Size > wRecNo_Count Then Begin
                              liNextSeqToLoad     := wRecNo_Count - csiBuffer_Size + 2;
                              wPreviousSeq        := arZones[bsBuffer.liCount].wSeq;
                              LoadBuffer(liNextSeqToLoad);
                              liZoneNumberOf1stRow := GetSeqPosInBuffer(wPreviousSeq) - byTableLength + 1;
                              liHighlightedZone    := GetSeqPosInBuffer(wPreviousSeq) + 1;
                            End
                            Else Begin
                              // A full buffer remains available
                              LoadBuffer(liNextSeqToLoad);
                            End; // If liNextSeqToLoad
                            DrawFullPage;
                          End
                          Else Begin
                            // No, end of valid zones, we stay where we are
                            // Display message like "we're already at the end" ??
                          End; // If arZones
                        End
                        Else Begin
                          // No, proceed normally inside active buffer
                          If liHighlightedZone < bsBuffer.liCount Then Begin
                            If liHighlightedZone < liZoneNumberOf1stRow+byTableLength Then Begin
                              Draw_Bar(liHighlightedZone, False);
                              liHighlightedZone := liHighlightedZone + 1;
                              Draw_Bar(liHighlightedZone, True);
                            End
                            Else If liHighlightedZone < bsBuffer.liCount Then Begin
                              liZoneNumberOf1stRow := liZoneNumberOf1stRow + 1;
                              liHighlightedZone    := liHighlightedZone  + 1;
                              DrawFullPage;
                            End; // If liHighlightedZone < liZoneNumberOf1stRow...
                          End; // If liHighlightedZone < bsBuffer.liCount...
                        End; // If (liHighlightedZone = bsBuffer.liCount) And (bsBuffer.liRemaining>0)
                      End; // Chr(80)

        Chr(81),
        Chr(77)     : Begin
                        // key PageDown Or Right
                        ClearSearch;

                        // If not already on last line of screen, move there as first step
                        If liHighlightedZone < liZoneNumberOf1stRow + byTableLength Then Begin
                          Draw_Bar(liHighlightedZone, False);
                          liHighlightedZone := liZoneNumberOf1stRow + byTableLength;
                          If liHighlightedZone > bsBuffer.liCount Then liHighlightedZone := bsBuffer.liCount;
                          Draw_Bar(liHighlightedZone, True);
                        End
                        Else If liHighlightedZone + byTableLength + 1 > bsBuffer.liCount Then Begin
                          // Buffer exhausted, we need to read more data from .dat files
                          wPreviousSeq    := arZones[liHighlightedZone].wSeq;
                          liNextSeqToLoad := wPreviousSeq + 1;
                          If liNextSeqToLoad > wRecNo_Count - csiBuffer_Size Then Begin
                            // End of data, position to load the last available full buffer (last slice of csiBuffer_Size zones)
                            liNextSeqToLoad   := wRecNo_Count - csiBuffer_Size + 1;
                            If liNextSeqToLoad < 1 Then liNextSeqToLoad := 1;
                            // Load only if new pos is <> previous pos (already at the end?)
                            If liNextSeqToLoad <> arZones[1].wSeq Then Begin
                              LoadBuffer(liNextSeqToLoad);
                            End; // If liNextSeqToLoad <> arZones[1].wSeq
                            liHighlightedzone := GetSeqPosInBuffer(wPreviousSeq) + byTableLength+1;
                            If liHighlightedZone > bsBuffer.liCount Then Begin
                              liHighlightedZone := bsBuffer.liCount;
                            End; // If liHighlightedZone > bsBuffer.liCount
                            liZoneNUmberOf1stRow := liHighlightedZone - byTableLength;
                            DrawFullPage;
                          End
                          Else Begin
                            // Normal buffer read
                            liNextSeqToLoad   := arZones[liHighlightedZone].wSeq + 1;
                            LoadBUffer(liNextSeqToLoad);
                            liZoneNumberOf1stRow := 1;
                            liHighlightedZone    := liZoneNumberOf1stRow + byTableLength;
                            DrawFullPage;
                          End; // if liNextSeqToLoad > wRecNo_Count
                        End
                        // Buffer not exhausted, we move in it
                        Else If liZoneNumberOf1stRow+byTableLength+1 < bsBuffer.liCount-byTableLength+1 Then Begin
                          // End of buffer, we move to last row
                          liZoneNumberOf1stRow := liZoneNumberOf1stRow + byTableLength + 1;
                          liHighlightedZone    := liHighlightedZone  + byTableLength + 1;
                          DrawFullPage;
                        End
                        Else Begin
                          // Normal move in buffer
                          liZoneNumberOf1stRow := bsBuffer.liCount - byTableLength;
                          liHighlightedZone    := bsBuffer.liCount;
                          DrawFullPage;
                        End; // If liZoneNumberOf1stRow+byTableLength...
                      End; // Chr(81 | 77)
      End; // Case

      cChar := #0; // Reset input
    End
//    Else If cChar=Chr(32) Then Begin
//      AddLog('************ dumping arZones **********')
//      For wCount := 1 To bsBuffer.liCount+2 Do Begin
//        AddLog('['+Int2Str(wCount)+'].sName='+arZones[wCount].sName+'/seq:'+Int2Str(arZones[wCount].wSeq))
//      End;
//      AddLog('******** done *********')
//    End
    Else If cChar >= Chr(32) And cChar <= Chr(128) Then Begin
      // Key is AlphaNum
      // Do search only if more data than fit on a unique page
      If wRecNo_Count > byTableLength Then Begin
        sSearch := sSearch+Upper(cChar);
        GotoXY (byPos_SearchX, byPos_SearchY);
        Write (sColor_Search_Prefix+Upper(sSearch));

        SearchWithString(sSearch)
      End; // If wRecNo_Count
    End // If cChar >= Chr(32)...
    Else Begin
      Case cChar Of
            Chr(26)       : Begin
                              // Help, CTRL-Z ?
                              DispFile (CfgTextPath + sHelpFile);
                              cChar := ReadKey;
                              DispFile (sDisplayFile);
                              DrawFullPage;
                            End; // Chr(26)
            Chr(13),
            Chr(27)       : Begin
                              // Key is ENTER or ESC
                              bDone := True;
                            End; // Chr(13)
            Chr(8)        : Begin
                            	// Key is ???
                              Delete (sSearch, Length(sSearch), 1);
          	                  GotoXY (byPos_SearchX, byPos_SearchY);
          	                  Write (sColor_Search_Prefix+Upper(sSearch)+' ');

                              SearchWithString(sSearch);
                            End; // Chr(8)
            Chr(1)        : Begin
                              // Search because the search string has changed ?
                              SearchWithString(sSearch);
                            End; // Chr(1)
            Chr(25)       : Begin
                              // Clear/Reset search string
                              ClearSearch;
                              LoadBuffer(1);
                              liZoneNumberOf1stRow := 1;
                              liHighlightedZone    := 1;
                              DrawFullPage;
                            End; // Chr(25)
      End; // Case cChar Of...
    End; // If IsArrow
  Until bDone; // Repeat

  GotoXY (1, 23);

  // ENTER key
  If cChar = Chr(13) Then Begin
      MenuCmd (sDefaultMenu, Int2Str(arZones[liHighlightedZone].liUniqueID));
  End;
End.
